#include <pfs.h>
#include <ImfHeader.h>
#include <ImfChannelList.h>
#include <ImfInputFile.h>
#include <ImfRgbaFile.h>
#include <ImfStringAttribute.h>
#include <ImfStandardAttributes.h>
#include <QByteArray>
// #include <stdio.h>
#include <iostream>
#include <stdlib.h>
using namespace Imf;
using namespace Imath;

using std::string;

static string escapeString( const string &src )
{
  int pos = 0;
  string ret = src;
  while( pos < (int)ret.size() ) {
    pos = ret.find( "\n", pos );
    if( pos == -1 ) break;
    ret.replace( pos, 1, "\\n" );
    pos+=2;
  }
  return ret;
}

#ifndef _WIN32
QByteArray readEXRfile (const char * filename) {
    char *pfshdr;
    size_t size;
    FILE *fpout=open_memstream (&pfshdr, &size);
#else
void readEXRfile (const char * filename,const char* outfilename) {
    FILE *fpout=fopen(outfilename,"wb");
#endif

    pfs::DOMIO pfsio;
    FILE *inputEXRfile=fopen(filename,"rb");
    InputFile file( filename );

    FrameBuffer frameBuffer;
    Box2i dw = file.header().dataWindow();
    int width  = dw.max.x - dw.min.x + 1;
    int height = dw.max.y - dw.min.y + 1;
    pfs::Frame *frame = pfsio.createFrame( width, height );
    const ChannelList &channels = file.header().channels();
    pfs::Channel *X, *Y, *Z;
      const Channel *rChannel = channels.findChannel( "R" );
      const Channel *gChannel = channels.findChannel( "G" );
      const Channel *bChannel = channels.findChannel( "B" );
      if( rChannel!=NULL && gChannel!=NULL && bChannel!=NULL ) {
//       qDebug("found RGB chans");
        frame->createXYZChannels( X, Y, Z );

        frameBuffer.insert( "R",				  // name
          Slice( FLOAT,			  // type
            (char *)X->getRawData(),
            sizeof(float),	  // xStride
            sizeof(float) * width,// yStride
            1, 1,			  // x/y sampling
            0.0));			  // fillValue

        frameBuffer.insert( "G",				  // name
          Slice( FLOAT,			  // type
            (char *)Y->getRawData(),
            sizeof(float),	  // xStride
            sizeof(float) * width,// yStride
            1, 1,			  // x/y sampling
            0.0));			  // fillValue

        frameBuffer.insert( "B",				  // name
          Slice( FLOAT,			  // type
            (char *)Z->getRawData(),
            sizeof(float),	  // xStride
            sizeof(float) * width,// yStride
            1, 1,			  // x/y sampling
            0.0)); // fillValue

      } //else qDebug("no rgb channels in exr");


      for( Header::ConstIterator it = file.header().begin();
           it != file.header().end(); it++ ) {
        const char *attribName = it.name();
        const char *colon = strstr( attribName, ":" );
        const StringAttribute *attrib =
          file.header().findTypedAttribute<StringAttribute>(attribName);

        if( attrib == NULL ) continue; // Skip if type is not String
        if( colon == NULL ) {   // frame tag
          frame->getTags()->setString( attribName, escapeString(attrib->value()).c_str() );
        } else {                // channel tag
          string channelName = string( attribName, colon-attribName );
          pfs::Channel *ch = frame->getChannel( channelName.c_str() );
          if( ch == NULL ) {
            continue;
          }
          ch->getTags()->setString(  colon+1, escapeString( attrib->value() ).c_str() );
        }
      }

    file.setFrameBuffer( frameBuffer );
    file.readPixels( dw.min.y, dw.max.y );

      // Rescale values if WhiteLuminance is present
      if( hasWhiteLuminance( file.header() ) ) {
        float scaleFactor = whiteLuminance( file.header() );
        int pixelCount = frame->getHeight()*frame->getWidth();
        for( int i = 0; i < pixelCount; i++ ) {
          (*X)(i) *= scaleFactor;
          (*Y)(i) *= scaleFactor;
          (*Z)(i) *= scaleFactor;
        }
//         const StringAttribute *relativeLum =
//           file.header().findTypedAttribute<StringAttribute>("RELATIVE_LUMINANCE");

        const char *luminanceTag = frame->getTags()->getString( "LUMINANCE" );
        if( luminanceTag == NULL )
          frame->getTags()->setString( "LUMINANCE", "ABSOLUTE" );
      }

    pfs::transformColorSpace( pfs::CS_RGB, X, Y, Z, pfs::CS_XYZ, X, Y, Z );
    frame->getTags()->setString( "FILE_NAME", filename );

    pfsio.writeFrame( frame, fpout );
    pfsio.freeFrame( frame );

    fclose(fpout);
    fclose(inputEXRfile);
#ifndef _WIN32
    QByteArray toreturn;
    toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
//     free(pfshdr); //since fromRawData is shallow copy, I cannot free it.
    return toreturn;
#endif
}
