/**
 * @file
 * @brief Writing QImage from PFS stream (which is a tonemapped LDR)
 * 
 * This file is a part of Qtpfsgui package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2006 Giuseppe Rota
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 * 
 * @author Giuseppe Rota <grota@users.sourceforge.net>
 *
 * $Id: pfsoutldrimage.cpp,v 1.0 2006/12/12 13:36:54
 */


#include <iostream>
#include <pfs.h>
#include <QImage>
#include <QByteArray>

static inline unsigned char clamp( const float v, const unsigned char minV, const unsigned char maxV )
{
    if( v < /*(float)*/minV ) return minV;
    if( v > /*(float)*/maxV ) return maxV;
    return (unsigned char)v;
}


#ifdef _WIN32
QImage* fromLDRPFStoPPM_QImage( const char* infilename, uchar * data) {
	FILE *inputfp=fopen(infilename,"rb");
#else
QImage* fromLDRPFStoPPM_QImage( QByteArray inqba, uchar * data) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
#endif
	if(data != NULL)
		delete [] data;

	pfs::DOMIO pfsio;
	pfs::Frame *frame = pfsio.readFrame( inputfp );
	assert( frame != NULL );

	pfs::Channel *X, *Y, *Z;
	frame->getXYZChannels( X, Y, Z );
	assert( X!=NULL && Y!=NULL && Z!=NULL );
	pfs::transformColorSpace( pfs::CS_XYZ, X, Y, Z, pfs::CS_SRGB, X, Y, Z );

	int width = X->getCols();
	int height =  X->getRows();
	data=new uchar[width*height*4]; //this will contain the image data: data must be 32-bit aligned, in Format: 0xffRRGGBB
	for( int y = 0; y < height; y++ ) { // For each row of the image
		for( int x = 0; x < width; x++ ) {
			*(data + 0 + (y*width+x)*4) = ( clamp( (*Z)( x, y )*255.f, 0, 255) );
			*(data + 1 + (y*width+x)*4) = ( clamp( (*Y)( x, y )*255.f, 0, 255) );
			*(data + 2 + (y*width+x)*4) = ( clamp( (*X)( x, y )*255.f, 0, 255) );
			*(data + 3 + (y*width+x)*4) = 0xff;
		}
	}

	pfsio.freeFrame( frame ); // would this delete the previous buffer?
	fclose(inputfp);
	QImage *toreturn=new QImage(data,width,height,QImage::Format_RGB32);
	return toreturn;
}
