/**
 * @brief Resize images in PFS stream
 * 
 * This file is a part of PFSTOOLS package.
 * ---------------------------------------------------------------------- 
 * Copyright (C) 2003,2004 Rafal Mantiuk and Grzegorz Krawczyk,
 *  Alexander Efremov
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ---------------------------------------------------------------------- 
 *
 * @author Alexander Efremov, <aefremov@mpi-sb.mpg.de>
 *
 * $Id: pfsrotate.cpp,v 1.1 2005/06/15 13:36:54 rafm Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pfs.h>
#include <sstream>
#include <QByteArray>

void rotateArray( const pfs::Array2D *in, pfs::Array2D *out, bool clockwise );

//FILE_IN = FILE_OUT IN PFSROTATE!
#ifdef _WIN32
void rotateFrame( const char* input, bool clock_wise ) {
	FILE *inputfp=fopen(input,"rb");
#else
QByteArray rotateFrame( QByteArray inqba, bool clock_wise ) {
	assert(inqba.size()!=0);
	FILE *inputfp=fmemopen(inqba.data(), (size_t)(inqba.size()), "r");
	//prepare "fake" FILE* where pfs library will write the output hdr to, in pfs stream format.
	char *pfshdr;
	size_t size;
	FILE * fpout=open_memstream (&pfshdr, &size);
#endif

    pfs::DOMIO pfsio;
    bool clockwise = clock_wise;
    int xSize = -1;
    int ySize = -1;

    pfs::Frame *resizedFrame = NULL;

    pfs::Frame *frame = pfsio.readFrame( inputfp );

	fclose(inputfp);
#ifdef _WIN32
	FILE *fpout=fopen(input,"wb");
#endif

    pfs::Channel *X, *Y, *Z;
    frame->getXYZChannels( X, Y, Z );

    xSize = frame->getHeight();
    ySize = frame->getWidth();
    resizedFrame = pfsio.createFrame( xSize, ySize );

    pfs::ChannelIterator *it = frame->getChannels();
    while( it->hasNext() ) {
      pfs::Channel *originalCh = it->getNext();
      pfs::Channel *newCh = resizedFrame->createChannel( originalCh->getName() );

      rotateArray( originalCh, newCh, clockwise );
    }

    pfs::copyTags( frame, resizedFrame );
    pfsio.writeFrame( resizedFrame, fpout );
    pfsio.freeFrame( frame );

    pfsio.freeFrame( resizedFrame ); // would this delete the previous buffer?

    fclose(fpout);
#ifndef _WIN32
    QByteArray toreturn;
    toreturn=QByteArray::fromRawData(pfshdr,size); //shallow copy
//     free(pfshdr); //since QByteArray::fromRawData is shallow copy, I cannot free it.
    return toreturn;
#endif
}

void rotateArray(const pfs::Array2D *in, pfs::Array2D *out, bool clockwise)
{
  int outRows = out->getRows();
  int outCols = out->getCols();

  for( int i=0; i<outCols; i++ )
    for( int j=0; j<outRows; j++ )
      if( clockwise )
        (*out)( i, j ) = (*in)( j, outCols - i - 1 );
      else
        (*out)( i, j ) = (*in)( outRows - j - 1, i );
}
