#ifndef  IMAGEHDRVIEWER_H
#define IMAGEHDRVIEWER_H
#include <QMainWindow>
#include <QImage>
#include <array2d.h>
#include <pfs.h>
#include <QComboBox>
#include <QPaintEvent>
#include <QPainter>
#include <QScrollBar>
#include <QScrollArea>
#include <QLabel>
#include <QToolBar>
#include "luminancerange_widget.h"

enum LumMappingMethod {
  MAP_LINEAR,
  MAP_GAMMA1_4,
  MAP_GAMMA1_8,
  MAP_GAMMA2_2,
  MAP_GAMMA2_6,
  MAP_LOGARITHMIC
};


class ImageMDIwindow : public QMainWindow {
	Q_OBJECT
public:
	ImageMDIwindow ( QWidget *parent);
	~ImageMDIwindow();
	LuminanceRangeWidget *lumRange;
	QToolBar *toolBar;
	double getScaleFactor();
	bool getFittingWin();
#ifdef _WIN32
public:
	void updateHDR(const char*);
#else
public:
	void updateHDR(QByteArray);
	QByteArray getHdrPfsStream();
private:
	QByteArray hdrpfsstream;
#endif

public slots:
	void updateRangeWindow();
	void setLumMappingMethod( int method );
	void zoomIn();
	void zoomOut();
	void fitToWindow(bool checked);
	void normalSize();
protected:
	QLabel *imageLabel;
	QScrollArea *scrollArea;
	QImage *image;
	void setRangeWindow( float min, float max );
	const pfs::Array2D *getPrimaryChannel();
	void updateImage();
private:
	void scaleImage(double);
	void adjustScrollBar(QScrollBar *scrollBar, double factor);
	QComboBox *mappingMethodCB;
	pfs::Array2D *workArea[3];
	pfs::Frame *pfsFrame;
	LumMappingMethod mappingMethod;
	float minValue;
	float maxValue;
	double scaleFactor;
	bool fittingwin;
};
#endif
