/**
 * @file tmo_reinhard02.h
 * @brief Tone map luminance channel using Reinhard02 model
 *
 * @author Grzegorz Krawczyk, <krawczyk@mpi-sb.mpg.de>
 *
 * $Id: tmo_reinhard02.h,v 1.2 2005/12/05 13:03:39 mantiuk Exp $
 */

#ifndef _tmo_reinhard02_h_
#define _tmo_reinhard02_h_

#include <QByteArray>
/*
 * @brief Photographic tone-reproduction
 *
 * @param Y input luminance
 * @param L output tonemapped intensities
 * @param use_scales true: local version, false: global version of TMO
 * @param key maps log average luminance to this value (default: 0.18)
 * @param phi sharpening parameter (defaults to 1 - no sharpening)
 * @param num number of scales to use in computation (default: 8)
 * @param low size in pixels of smallest scale (should be kept at 1)
 * @param high size in pixels of largest scale (default 1.6^8 = 43)
 */
extern "C" {
void tmo_reinhard02(const pfs::Array2D *Y, pfs::Array2D *L, 
  bool use_scales, float key, float phi, 
  int num, int low, int high, bool temporal_coherent );
#ifdef _WIN32
QByteArray pfstmo_reinhard02 (const char* input, const char* output, float _key, float _phi, int _num, int _low, int _high, bool _use_scales );
#else
QByteArray pfstmo_reinhard02 (QByteArray inqba, float _key, float _phi, int _num, int _low, int _high, bool _use_scales );
#endif
}
#endif /* _tmo_reinhard02_h_ */
